# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import thermosteam as tmo
from numpy.testing import assert_allclose

def test_chemical_ID_parsing():
    H2O = tmo.Chemical('H2O,g')
    assert H2O.ID == 'H2O'
    assert H2O.locked_state == 'g'
    Steam = tmo.Chemical('Steam,g', search_ID='H2O')
    assert Steam.ID == 'Steam'
    assert Steam.locked_state == 'g'
    chemicals = tmo.Chemicals(['H2O', 'Ethanol', 'O2,g', 'CO2,g'])
    assert [i.ID for i in chemicals] == ['H2O', 'Ethanol', 'O2', 'CO2']
    assert [i.locked_state for i in chemicals] == [None, None, 'g', 'g']
    with pytest.raises(ValueError):
        tmo.Chemical('CO2,v')
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# A significant portion of this module originates from:
# Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
# 
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/thermo/blob/master/LICENSE.txt for details.
"""
"""
import numpy as np
from warnings import warn
from .unifac import DOUFSG, DOUFIP2016, UFIP, UFSG, NISTUFSG, NISTUFIP
from numba import njit, prange
from .ideal import ideal
import thermosteam as tmo
from thermo.interaction_parameters import IPDB
from fluids.constants import R_inv
from thermo import eos_mix
import numpy as np

__all__ = ('ActivityCoefficients',
           'IdealActivityCoefficients',
           'GroupActivityCoefficients',
           'DortmundActivityCoefficients',
           'UNIFACActivityCoefficients',
           'NISTActivityCoefficients')

# %% Utilities

def chemgroup_array(chemgroups, index):
    M = len(chemgroups)
    N = len(index)
    array = np.zeros((M, N))
    for i, groups in enumerate(chemgroups):
        for group, count in groups.items():
            array[i, index[group]] = count
    return array

@njit(cache=True)
def group_activity_coefficients(x, chemgroups, loggammacs,
                                Qs, psis, cQfs, gpsis):
    weighted_counts = chemgroups.transpose() @ x
    Q_fractions = Qs * weighted_counts 
    Q_fractions /= Q_fractions.sum()
    Q_psis = psis * Q_fractions
    sum1 = Q_psis.sum(1)
    sum2 = -(psis.transpose() / sum1) @ Q_fractions
    loggamma_groups = Qs * (1. - np.log(sum1) + sum2)
    sum1 = cQfs @ gpsis.transpose()
    sum1 = np.where(sum1==0, 1., sum1)
    fracs = - cQfs / sum1
    sum2 = fracs @ gpsis
    chem_loggamma_groups = Qs*(1. - np.log(sum1) + sum2)
    loggammars = ((loggamma_groups - chem_loggamma_groups) * chemgroups).sum(1)
    return np.exp(loggammacs + loggammars)

def get_interaction(all_interactions, i, j, no_interaction):
    if i==j:
        return no_interaction
    try:
        return all_interactions[i][j]
    except:
        return no_interaction

def get_chemgroups(chemicals, field):
    getfield = getattr
    chemgroups = []
    index = []
    for i, chemical in enumerate(chemicals): 
        group = getfield(chemical, field)
        if group:
            chemgroups.append(group)
            index.append(True)
        else:
            index.append(False)
            warn(f"{chemical} has no defined {field} groups; "
                  "functional group interactions are ignored",
                  RuntimeWarning, stacklevel=3)
    return np.where(index)[0], chemgroups

@njit(cache=True)
def loggammacs_UNIFAC(qs, rs, x):
    r_net = np.dot(x, rs)
    q_net = np.dot(x, qs)
    Vs = rs/r_net
    Fs = qs/q_net
    Vs_over_Fs = Vs/Fs
    return 1. - Vs - np.log(Vs) - 5.*qs*(1. - Vs_over_Fs + np.log(Vs_over_Fs))

@njit(cache=True)
def loggammacs_modified_UNIFAC(qs, rs, x):
    r_net = np.dot(x, rs)
    q_net = np.dot(x, qs)
    rs_p = rs**0.75
    r_pnet = np.dot(rs_p, x)
    Vs = rs/r_net
    Fs = qs/q_net
    Vs_over_Fs = Vs/Fs
    Vs_p = rs_p/r_pnet
    return 1. - Vs_p + np.log(Vs_p) - 5.*qs*(1. - Vs_over_Fs + np.log(Vs_over_Fs))

@njit(cache=True)
def psi_modified_UNIFAC(T, abc):
    abc[:, :, 0] /= T
    abc[:, :, 2] *= T
    return np.exp(-abc.sum(2)) 

@njit(cache=True)
def psi_UNIFAC(T, a):
    return np.exp(-a/T)

@njit(cache=True)
def fill_group_psis(group_psis, psis, group_mask):
    M = psis.shape[0]
    N = psis.shape[1]
    for i in range(M):
        for j in range(N):
            if group_mask[i, j]: 
                group_psis[i, j] = psis[i, j]
            else:
                group_psis[i, j] = 0.

@njit(cache=True)
def gamma_UNIFAC(x, T, interactions, 
                 group_psis, group_mask, qs, rs, Qs,
                 chemgroups, chem_Qfractions, index):
    N_chemicals = index.size
    gamma = np.ones(x.size)
    if N_chemicals > 1:
        interactions = interactions.copy()
        x_sub = np.ones(N_chemicals)
        for i, j in enumerate(index): x[j] = x_sub[i]
        xsum = x_sub.sum()
        if xsum != 0: 
            x_sub /= xsum
            psis = psi_UNIFAC(T, interactions)
            fill_group_psis(group_psis, psis, group_mask)
            gamma_sub = group_activity_coefficients(x_sub, chemgroups,
                                        loggammacs_UNIFAC(qs, rs, x_sub),
                                        Qs, psis,
                                        chem_Qfractions,
                                        group_psis)
        for i, j in enumerate(index): 
            value = gamma_sub[i]
            if np.isnan(value): continue
            gamma[j] = value
    return gamma

@njit(cache=True)
def gamma_modified_UNIFAC(x, T, interactions, 
                   group_psis, group_mask, qs, rs, Qs,
                   chemgroups, chem_Qfractions, index):
    N_chemicals = index.size
    gamma = np.ones(x.size)
    if N_chemicals > 1:
        interactions = interactions.copy()
        x_sub = np.ones(N_chemicals)
        for i, j in enumerate(index): x_sub[i] = x[j]
        xsum = x_sub.sum()
        if xsum:
            x_sub /= xsum
            psis = psi_modified_UNIFAC(T, interactions)
            fill_group_psis(group_psis, psis, group_mask)
            gamma_sub = group_activity_coefficients(
                x_sub, chemgroups,
                loggammacs_modified_UNIFAC(qs, rs, x_sub),
                Qs, psis,
                chem_Qfractions,
                group_psis
            )
        for i, j in enumerate(index): 
            value = gamma_sub[i]
            if np.isnan(value): continue
            gamma[j] = value
    return gamma
    
    
# %% Activity Coefficients

class ActivityCoefficients:
    """
    Abstract class for the estimation of activity coefficients. 
    Non-abstract subclasses should implement the following methods:
        
    __init__(self, chemicals: Iterable[:class:`~thermosteam.Chemical`]):
        Should use pure component data from chemicals to setup future 
        calculations of activity coefficients.
    
    __call__(self, x: 1d array, T: float):
        Should accept an array of liquid molar compositions `x`, and
        temperature `T` (in Kelvin), and return an array of activity 
        coefficients. Note that the molar compositions must be in the same 
        order as the chemicals defined when creating the ActivityCoefficients
        object.
    
    """
    __slots__ = ('_chemicals',)
    
    @property
    def chemicals(self):
        """tuple[Chemical] All chemicals involved in the calculation of activity coefficients."""
        return self._chemicals
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.chemicals])
        return f"{type(self).__name__}([{chemicals}])"

@ideal
class IdealActivityCoefficients(ActivityCoefficients):
    """
    Create an IdealActivityCoefficients object that estimates all activity 
    coefficients to be 1 when called with a composition and a temperature (K).
    
    Parameters
    ----------
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ()
    
    def __init__(self, chemicals):
        self._chemicals = tuple(chemicals)
    
    def __call__(self, xs, T):
        return np.ones(len(xs))
    

class GroupActivityCoefficients(ActivityCoefficients):
    """
    Abstract class for the estimation of activity coefficients using
    group contribution methods.
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_rs', '_qs', '_Qs','_chemgroups',
                 '_group_psis',  '_chem_Qfractions',
                 '_group_mask', '_interactions',
                 '_chemicals', '_index')
    
    def __new__(cls, chemicals):
        chemicals = tuple(chemicals)
        if chemicals in cls._cached:
            return cls._cached[chemicals]
        else:
            self = super().__new__(cls)
        index, chemgroups = get_chemgroups(chemicals, self.group_name)
        if len(index) <= 1: return IdealActivityCoefficients(chemicals)
        self._index = index
        all_groups = set()
        for groups in chemgroups: all_groups.update(groups)
        index = {group:i for i,group in enumerate(all_groups)}
        chemgroups = chemgroup_array(chemgroups, index)
        all_subgroups = self.all_subgroups
        subgroups = [all_subgroups[i] for i in all_groups]
        main_group_ids = [i.main_group_id for i in subgroups]
        self._Qs = Qs = np.array([i.Q for i in subgroups])
        Rs = np.array([i.R for i in subgroups])
        self._rs = chemgroups @ Rs
        self._qs = chemgroups @ Qs
        self._chemgroups = chemgroups
        chem_Qs = Qs * chemgroups
        self._chem_Qfractions = cQfs = chem_Qs/chem_Qs.sum(1, keepdims=True)
        all_interactions = self.all_interactions
        N_groups = len(all_groups)
        group_shape = (N_groups, N_groups)
        no_interaction = self._no_interaction
        self._interactions = np.array(
            [[get_interaction(all_interactions, i, j, no_interaction)
              for i in main_group_ids]
             for j in main_group_ids])
        # Psis array with only symmetrically available groups
        self._group_psis = np.zeros(group_shape, dtype=float)
        # Make mask for retrieving symmetrically available groups
        rowindex = np.arange(N_groups, dtype=int)
        indices = [rowindex[rowmask] for rowmask in cQfs != 0]
        self._group_mask = group_mask = np.zeros(group_shape, dtype=bool)
        for index in indices:
            for i in index:
                group_mask[i, index] = True
        self._cached[chemicals] = self
        self._chemicals = chemicals
        return self
    
    def __reduce__(self):
        return type(self), (self.chemicals,)
    
    @property
    def args(self):
        return (self._interactions, 
                self._group_psis, self._group_mask,
                self._qs, self._rs, self._Qs,
                self._chemgroups, self._chem_Qfractions, 
                self._index)
    
    def activity_coefficients(self, x, T):
        """
        Return activity coefficients of chemicals with defined functional groups.
        
        Parameters
        ----------
        x : array_like
            Molar fractions
        T : float
            Temperature [K]
        
        """
        psis = self.psi(T, self._interactions.copy())
        self._group_psis[self._group_mask] =  psis[self._group_mask]
        return group_activity_coefficients(x, self._chemgroups,
                                           self.loggammacs(self._qs, self._rs, x),
                                           self._Qs, psis,
                                           self._chem_Qfractions,
                                           self._group_psis)
    
    def __call__(self, x, T):
        """
        Return activity coefficients.
        
        Parameters
        ----------
        x : array_like
            Molar fractions
        T : float
            Temperature [K]
        
        Notes
        -----
        Activity coefficients of chemicals with missing groups are default to 1.
        
        """
        x = np.asarray(x, float)
        return self.f(x, T, *self.args)
    
    
class UNIFACActivityCoefficients(GroupActivityCoefficients):
    """
    Create a UNIFACActivityCoefficients that estimates activity coefficients 
    using the UNIFAC group contribution method when called with a composition
    and a temperature (K).
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    all_subgroups = UFSG
    all_interactions = UFIP
    group_name = 'UNIFAC'
    _no_interaction = 0.
    _cached = {}
    
    @property
    def f(self):
        return gamma_UNIFAC
    
    @property
    def loggammacs(self):
        return loggammacs_UNIFAC
    
    @property
    def psi(self):
        return psi_UNIFAC


class DortmundActivityCoefficients(GroupActivityCoefficients):
    """
    Create a DortmundActivityCoefficients that estimates activity coefficients
    using the Dortmund UNIFAC group contribution method when called with a 
    composition and a temperature (K).
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
    >>> Gamma = tmo.equilibrium.DortmundActivityCoefficients(chemicals)
    >>> composition = [0.5, 0.5]
    >>> T = 350.                                
    >>> Gamma(composition, T)
    array([1.475, 1.242])
    
    >>> chemicals = tmo.Chemicals(['Dodecane', 'Tridecane'], cache=True)
    >>> Gamma = tmo.equilibrium.DortmundActivityCoefficients(chemicals)
    >>> # Note how both hydrocarbons have similar lenghts and structure,
    >>> # so activities should be very close
    >>> Gamma([0.5, 0.5], 350.) 
    array([1., 1.])
    
    >>> chemicals = tmo.Chemicals(['Water', 'O2'], cache=True)
    >>> Gamma = tmo.equilibrium.DortmundActivityCoefficients(chemicals)
    >>> # The following warning is issued because O2 does not have Dortmund groups
    >>> # RuntimeWarning: O2 has no defined Dortmund groups; 
    >>> # functional group interactions are ignored
    >>> Gamma([0.5, 0.5], 350.) 
    array([1., 1.])
    
    """
    __slots__ = ()
    all_subgroups = DOUFSG
    all_interactions = DOUFIP2016
    group_name = 'Dortmund'
    _no_interaction = np.array([0., 0., 0.])
    _cached = {}
    
    @property
    def f(self):
        return gamma_modified_UNIFAC
    
    @property
    def loggammacs(self):
        return loggammacs_modified_UNIFAC
    
    @property
    def psi(self):
        return psi_modified_UNIFAC
    
    
class NISTActivityCoefficients(GroupActivityCoefficients):
    """
    Create a NISTActivityCoefficients that estimates activity coefficients
    using the NIST-modified UNIFAC group contribution method when called with a 
    composition and a temperature (K).
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    Examples
    --------
    >>> import thermosteam as tmo
    >>> Water, Ethanol = chemicals = tmo.Chemicals(['Water', 'Ethanol'], cache=True)
    >>> Ethanol.NIST.set_group_counts_by_name({'CH3':1, 'CH2':1, 'OH prim':1})
    >>> Water.NIST.set_group_counts_by_name({'H2O':1})
    >>> Gamma = tmo.equilibrium.NISTActivityCoefficients(chemicals)
    >>> composition = [0.5, 0.5]
    >>> T = 350.                                
    >>> Gamma(composition, T)
    array([1.479, 1.238])
    
    """
    __slots__ = ()
    all_subgroups = NISTUFSG
    all_interactions = NISTUFIP
    group_name = 'NIST'
    _no_interaction = np.array([0., 0., 0.])
    _cached = {}
    
    f = DortmundActivityCoefficients.f
    
    @property
    def loggammacs(self):
        return loggammacs_modified_UNIFAC
    
    @property
    def psi(self):
        return psi_modified_UNIFAC


class GCEOSActivityCoefficients(ActivityCoefficients):
    """
    Abstract class for estimating fugacity coefficients using a generic cubic equation of state 
    when called with composition, temperature (K), and pressure (Pa).
    
    Parameters
    ----------
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_chemicals', '_eos')
    EOS = None # type[GCEOSMIX] Subclasses must implement this attribute.
    chemsep_db = None # Optional[str] Name of chemsep data base for interaction parameters.
    cache = None # [dict] Subclasses must implement this attribute.
    
    def __new__(cls, chemicals):
        chemicals = tuple(chemicals)
        cache = cls.cache
        if chemicals in cache:
            return cache[chemicals]
        else:
            self = super().__new__(cls)
            self._chemicals = chemicals
            cache[chemicals] = self
            return self
    
    @classmethod
    def subclass(cls, EOS, name=None):
        if name is None: name = EOS.__name__[:-3] + 'ActivityCoefficients'
        return type(name, (cls,), dict(EOS=EOS, cache={}))
    
    @property
    def chemicals(self):
        """tuple[Chemical] All chemicals involved in the calculation of fugacity coefficients."""
        return self._chemicals
    
    def eos(self, zs, T, P):
        if zs.__class__ is np.ndarray: zs = [float(i) for i in zs]
        try:
            self._eos = eos = self._eos.to_TP_zs_fast(T=T, P=P, zs=zs, only_l=True, full_alphas=False)
        except:
            data = tmo.ChemicalData(self.chemicals)
            if self.chemsep_db is None:
                kijs = None
            else:
                try:
                    kijs = IPDB.get_ip_asymmetric_matrix(self.chemsep_db, data.CASs, 'kij')
                except:
                    kijs = None
            self._eos = eos = self.EOS(
                zs=zs, T=T, P=P, Tcs=data.Tcs, Pcs=data.Pcs, omegas=data.omegas, kijs=kijs,
                only_l=True
            )
        return eos
    
    def __call__(self, x, T, P=101325):
        eos = self.eos(x, T, P)
        try:
            log_phis = np.array(eos.dlnphi_dns(eos.Z_l)) + eos.G_dep_l * R_inv / T
            return np.exp(log_phis)
        except:
            return np.ones(len(x))
    f = __call__
    args = ()
        
dct = globals()    
activity_coefficient_classes = [UNIFACActivityCoefficients, DortmundActivityCoefficients, NISTActivityCoefficients]
clsnames = []
for name in ('PRMIX', 'SRKMIX', 'PR78MIX', 'VDWMIX', 'PRSVMIX',
             'PRSV2MIX', 'TWUPRMIX', 'TWUSRKMIX', 'APISRKMIX', 'IGMIX', 'RKMIX',
             'PRMIXTranslatedConsistent', 'PRMIXTranslatedPPJP', 'PRMIXTranslated',
             'SRKMIXTranslatedConsistent', 'PSRK', 'MSRKMIXTranslated',
             'SRKMIXTranslated'):
    cls = GCEOSActivityCoefficients.subclass(getattr(eos_mix, name))
    clsname = cls.__name__
    clsnames.append(clsname)
    activity_coefficient_classes.append(cls)
    dct[clsname] = cls

dct['PRActivityCoefficients'].chemsep_db = 'ChemSep PR'
__all__ = (*__all__, *clsnames, 'activity_coefficient_classes')
del dct, clsnames


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('PoyintingCorrectionFactors',
           'MockPoyintingCorrectionFactors',
           'IdealGasPoyintingCorrectionFactors',
           'poyinting_correction_factor_classes')
from ..constants import R
from numba import njit
import numpy as np

class PoyintingCorrectionFactors:
    """Abstract class for the estimation of Poyinting correction factors.
    Non-abstract subclasses should implement the following methods:
        
    __init__(self, chemicals: Iterable[:class:`~thermosteam.Chemical`]):
        Should use pure component data from chemicals to setup future 
        calculations of Poyinting correction factors.
    
    __call__(self, T: float, P: float):
        Should accept the temperature `T` (in Kelvin) and pressure `P` (in Pascal),
        and return an array of Poyinting correction factors. 
        
    """
    __slots__ = ()
    
    @property
    def chemicals(self):
        return self._chemicals
    @chemicals.setter
    def chemicals(self, chemicals):
        self._chemicals = tuple(chemicals)
    
    def __init__(self, chemicals):
        self.chemicals = chemicals
    
    def __reduce__(self):
        return type(self), (self.chemicals,)
    
    def __repr__(self):
        chemicals = ", ".join([i.ID for i in self.chemicals])
        return f"<{type(self).__name__}([{chemicals}])>"


class MockPoyintingCorrectionFactors(PoyintingCorrectionFactors):
    """Create a MockPoyintingCorrectionFactors object 
    that estimates all poyinting correction factors to be 1 when
    called with a temperature (K) and pressure (Pa).
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_chemicals',)
    
    def __call__(self, T, P, Psats=None):
        return 1.


@njit(cache=True)
def ideal_gas_poyinting_correction_factors(T, P, vls, Psats):
    dP = P - Psats
    dP[dP < 0.] = 0.
    return np.exp(vls / (R * T) * dP)

class IdealGasPoyintingCorrectionFactors(PoyintingCorrectionFactors):
    """Create an IdealGasPoyintingCorrectionFactors object that estimates 
    poyinting correction factors assuming ideal gas when called with 
    a temperature (K) and pressure (Pa).
    
    Parameters
    ----------
    
    chemicals : Iterable[:class:`~thermosteam.Chemical`]
    
    """
    __slots__ = ('_chemicals',)

    def __call__(self, T, P, Psats=None):
        vls = []
        chemicals = []
        index = []
        for i, chemical in enumerate(self._chemicals):
            try: vls.append(chemical.V('l', T, P))
            except: continue
            chemicals.append(chemical)
            index.append(i)
        vls = np.array(vls, dtype=float)
        if Psats is None:
            Psats = np.array([i.Psat(T) for i in chemicals], dtype=float)
        else:
            Psats = Psats[index]
        pcf = np.ones(len(self._chemicals))
        pcf[index] = ideal_gas_poyinting_correction_factors(T, P, vls, Psats)
        return pcf
        
poyinting_correction_factor_classes = [MockPoyintingCorrectionFactors, IdealGasPoyintingCorrectionFactors]


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('repr_IDs_data', 'repr_kwargs', 'repr_kwarg', 'repr_couples',
           'repr_listed_values', 'repr_obj')
    
def repr_IDs_data(IDs, data, dlim=", ", start=None): # Used for Indexer and Stream representation
    rep = dlim.join([f"{ID}={i:.4g}" for ID, i in zip(IDs, data) if i])    
    if not rep:
        return ''
    else:
        return (dlim if start is None else start) + rep

def repr_kwargs(kwargs, dlim=", ", start=None):
    """
    Represent key word arguments.

    Parameters
    ----------
    kwargs : dict[str: Any]
        Key workd arguments.
    dlim : str, optional
        Delimiter. The default is ", ".
    start : str, optional
        Start of return value. Defaults to delimiter value.

    Examples
    --------
    >>> repr_kwargs({'a': 1, 'b': 2})
    ', a=1, b=2'
    
    >>> repr_kwargs({'a': 1, 'b': 2}, start="")
    'a=1, b=2'
    
    """
    if kwargs:
        start = dlim if start is None else ""
        return start + dlim.join([f"{key}={value}" for key, value in kwargs.items()])
    else:
        return ""

def repr_args(args, dlim=", ", start=None):
    """
    Represent arguments.

    Parameters
    ----------
    args : Iterable[Any]
        Arguments.
    dlim : str, optional
        Delimiter. The default is ", ".
    start : str, optional
        Start of return value. Defaults to delimiter value.

    Examples
    --------
    >>> repr_args([1, 2])
    ', 1, 2'
    
    >>> repr_args([1, 2], start="")
    '1, 2'
    
    """
    if args:
        return (start if start is not None else dlim) + dlim.join([repr(value) for value in args])
    else:
        return ""

def repr_kwarg(name, value, spec=None, dlim=", "): 
    """
    Represent parameter for class __repr__ method.

    Parameters
    ----------
    name : str
        Name of parameter.
    value : Any
        Value being represented.
    spec : str, optional
        DESCRIPTION. The default is None.
    dlim : str, optional
        Delimiter. The default is ", ".

    Examples
    --------
    >>> repr_kwarg('price', 0.12324, spec='.2f', dlim=", ")
    ', price=0.12'
    
    """
    if value:
        if spec: value = format(value, spec)
        else: value = repr(value)
        return dlim + f"{name}={value}"
    else:
        return ""

def repr_couples(dlim, IDs, data):  # Used for Indexer and Stream representation
    return dlim.join([f"('{ID}', {i:.4g})" for ID, i in zip(IDs, data) if i])

def repr_listed_values(values):
    """
    Represent values for messages.
    
    Examples
    --------
    >>> repr_listed_values(['green', 'blue', 'yellow'])
    'green, blue and yellow'
    
    """
    *values, last = values 
    if values:
        return ", ".join(values) + ' and ' + last
    else:
        return last

def repr_obj(obj): 
    """
    Represent object.
    
    Examples
    --------
    >>> class Data:
    ...     def __init__(self, a=1, b=2):
    ...         self.a = a
    ...         self.b = b
    >>> repr_obj(Data())
    'Data(a=1, b=2)'
    
    """
    return f"{type(obj).__name__}({repr_kwargs(obj.__dict__, start='')})"







# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo

__all__ = ('chemicals_user',)

def chemicals_user(cls):
    cls._load_chemicals = _load_chemicals
    cls.chemicals = chemicals
    return cls

@property
def chemicals(self):
    return self._chemicals

def _load_chemicals(self, chemicals):
    self._chemicals = chemicals = tmo.settings.get_default_chemicals(chemicals)
    return chemicals


# -*- coding: utf-8 -*-
"""
"""
from chemicals.lennard_jones import Stockmayer, molecular_diameter
from thermo.unifac import UNIFAC_RQ, Van_der_Waals_volume, Van_der_Waals_area
from chemicals.utils import Parachor
from collections.abc import Iterable
import thermosteam as tmo

class CachedProperty:
    __slots__ = ('fget', 'key')
    
    def __new__(cls, fget):
        self = super().__new__(cls)
        self.fget = fget
        self.key = fget.__name__
        return self
    
    def __get__(self, data, _):
        cache = data.cache
        key = self.key
        if key in cache:
            value = cache[key]
        else:
            cache[key] = value = self.fget(data)
        return value


class ChemicalData:
    """
    Create a ChemicalData object that works as thermo ChemicalPropertiesPackage
    and ChemicalConstantsPackage objects.
    """
    __slots__ = ('N', 'chemicals', 'index', 'cmps', 'cache')
    
    def __init__(self, chemicals):
        if isinstance(chemicals, tmo.CompiledChemicals):
            self.chemicals = chemicals.tuple
            self.index = chemicals._index
            self.N = chemicals.size
        else:
            self.chemicals = tuple(chemicals)
            self.N = len(chemicals)
        self.cmps = range(self.N)
        self.cache = {}
    
    def __iter__(self):
        return self.chemicals.__iter__()
    
    def __getitem__(self, IDs):
        chemicals = self.chemicals
        try:
            index = self.index
        except:
            self.index = index = {}
            repeated_names = set()
            names = set()
            all_names_list = []
            for i in chemicals:
                if not i.iupac_name: i.iupac_name = ()
                all_names = set([*i.iupac_name, *i.aliases, i.common_name, i.formula])
                all_names_list.append(all_names)
                for name in all_names:
                    if not name: continue
                    if name in names:
                        repeated_names.add(name)
                    else:
                        names.add(name)
            for all_names, i in zip(all_names_list, chemicals):
                ID = i.ID
                for name in all_names:
                    if name and name not in repeated_names:
                        index[name] = index[ID]
        if isinstance(IDs, str):
            return chemicals[index[IDs]]
        elif isinstance(IDs, Iterable):
            return ChemicalData([chemicals[index[i]] for i in IDs])
        else:
            raise KeyError("key must be a string or a tuple, not a '{type(IDs).__name__}' object")
    
    @CachedProperty
    def EnthalpyVaporizations(self):
        return [i.Hvap for i in self.chemicals]
    @CachedProperty
    def VaporPressures(self):
        return [i.Psat for i in self.chemicals]
    @CachedProperty
    def SublimationPressures(self):
        return [i.Psub for i in self.chemicals]
    @CachedProperty
    def EnthalpySublimations(self):
        return [i.Hsub for i in self.chemicals]
    @CachedProperty
    def SurfaceTensions(self):
        return [i.sigma for i in self.chemicals]
    @CachedProperty
    def PermittivityLiquids(self):
        return [i.epsilon for i in self.chemicals]
    @CachedProperty
    def CASs(self):
        return [i.CAS for i in self.chemicals]
    @CachedProperty
    def MWs(self):
        return [i.MW for i in self.chemicals]
    @CachedProperty
    def Tms(self):
        return [i.Tm for i in self.chemicals]
    @CachedProperty
    def Tbs(self):
        return [i.Tb for i in self.chemicals]
    @CachedProperty
    def Tcs(self):
        return [i.Tc for i in self.chemicals]
    @CachedProperty
    def Pcs(self):
        return [i.Pc for i in self.chemicals]
    @CachedProperty
    def Vcs(self):
        return [i.Vc for i in self.chemicals]
    @CachedProperty
    def omegas(self):
        return [i.omega for i in self.chemicals]
    @CachedProperty
    def Zcs(self):
        return [i.Zc for i in self.chemicals]
    @CachedProperty
    def Hfus_Tms(self):
        return [i.Hfus for i in self.chemicals]   
    @CachedProperty
    def Tts(self):
        return [i.Tt for i in self.chemicals]
    @CachedProperty
    def Pts(self):
        return [i.Pt for i in self.chemicals]
    @CachedProperty
    def dipoles(self):
        return [i.dipole for i in self.chemicals]
    @CachedProperty
    def charges(self):
        return [i.charge for i in self.chemicals]
    @CachedProperty
    def UNIFAC_groups(self):
        return [i.UNIFAC for i in self.chemicals]
    @CachedProperty
    def UNIFAC_Dortmund_groups(self):
        return [i.Dortmund for i in self.chemicals]
    @CachedProperty
    def PSRK_groups(self):
        return [i.PSRK for i in self.chemicals]
    @CachedProperty
    def similarity_variables(self):
        return [i.similarity_variable for i in self.chemicals]
    @CachedProperty
    def StielPolars(self):
        return [i.Stiel_Polar for i in self.chemicals]
    @CachedProperty
    def VolumeGases(self):
        return [i.V if i._locked_state else i.V.g for i in self.chemicals]
    @CachedProperty
    def VolumeLiquids(self):
        return [i.V if i._locked_state else i.V.l for i in self.chemicals]
    @CachedProperty
    def VolumeSolids(self):
        return [i.V if i._locked_state else i.V.s for i in self.chemicals]
    @CachedProperty
    def HeatCapacityGases(self):
        return [i.Cn if i._locked_state else i.Cn.g for i in self.chemicals]
    @CachedProperty
    def HeatCapacityLiquids(self):
        return [i.Cn if i._locked_state else i.Cn.l for i in self.chemicals]
    @CachedProperty
    def HeatCapacitySolids(self):
        return [i.Cn if i._locked_state else i.Cn.s for i in self.chemicals]
    @CachedProperty
    def ViscosityGases(self):
        return [i.mu if i._locked_state else i.mu.g for i in self.chemicals]
    @CachedProperty
    def ViscosityLiquids(self):
        return [i.mu if i._locked_state else i.mu.l for i in self.chemicals]
    @CachedProperty
    def ThermalConductivityGases(self):
        return [i.k if i._locked_state else i.k.g for i in self.chemicals]
    @CachedProperty
    def ThermalConductivityLiquids(self):
        return [i.k if i._locked_state else i.k.l for i in self.chemicals]
    @CachedProperty
    def rhocs(self):
        return [None if Vc is None else 1.0/Vc for Vc in self.chemicals.Vcs]
    @CachedProperty
    def rhocs_mass(self):
        return [None if None in (rhoc, MW) else rhoc * 1e-3 * MW
                for rhoc, MW in zip(self.rhocs, self.MWs)]
    @CachedProperty
    def Hfus_Tms_mass(self):
        return [None if None in (Hfus, MW) else Hfus * 1000.0 / MW
                for Hfus, MW in zip(self.Hfus_Tms, self.MWs)]
    @CachedProperty
    def Hvap_Tbs(self):
        return [Hvap(Tb) for Hvap, Tb in zip(self.EnthalpyVaporizations, self.Tbs)]
    @CachedProperty
    def Hvap_Tbs_mass(self):
        return [None if None in (Hvap, MW) else Hvap*1000.0/MW
                for Hvap, MW in zip(self.Hvap_Tbs, self.Tbs)]
    @CachedProperty
    def Vmg_STPs(self):
        return [i.TP_dependent_property(298.15, 101325.0)
                for i in self.chemicals.VolumeGases]
    @CachedProperty
    def rhog_STPs(self):
        return [1.0/V if V is not None else None for V in self.chemicals.Vmg_STPs]
    @CachedProperty
    def rhog_STPs_mass(self):
        return [1e-3*MW/V if V is not None else None for V, MW in zip(self.Vmg_STPs, self.MWs)]
    @CachedProperty
    def Vml_Tbs(self):
        return [None if Tb is None else i.T_dependent_property(Tb) for i, Tb in zip(self.VolumeLiquids, self.Tbs)]
    @CachedProperty
    def Vml_Tms(self):
        return [None if Tm is None else i.T_dependent_property(Tm)
                for i, Tm in zip(self.VolumeLiquids, self.Tms)]
    @CachedProperty
    def Vml_STPs(self):
        return [i.T_dependent_property(298.15) for i in self.chemicals.VolumeLiquids]
    @CachedProperty
    def Vml_60Fs(self):
        return [i.T_dependent_property(288.7055555555555) for i in self.chemicals.VolumeLiquids]
    @CachedProperty
    def rhol_STPs(self):
        return [1.0/V if V is not None else None for V in self.chemicals.Vml_STPs]
    @CachedProperty
    def rhol_60Fs(self):
        return [1.0/V if V is not None else None for V in self.chemicals.Vml_60Fs]
    @CachedProperty
    def rhol_STPs_mass(self):
        return [1e-3*MW/V if V is not None else None for V, MW in zip(self.Vml_STPs, self.MWs)]
    @CachedProperty
    def rhol_60Fs_mass(self):
        return [1e-3*MW/V if V is not None else None for V, MW in zip(self.Vml_60Fs, self.MWs)]
    @CachedProperty
    def Hsub_Tts(self):
        return [None if Tt is None else i(Tt)
                for i, Tt in zip(self.EnthalpySublimations, self.Tts)]
    @CachedProperty
    def Hsub_Tts_mass(self):
        return [Hsub*1000.0/MW if Hsub is not None else None for Hsub, MW in zip(self.Hsub_Tts, self.MWs)]
    @CachedProperty
    def Stockmayers(self):
        Tms = self.Tms
        Tbs = self.Tbs
        Tcs = self.Tcs
        Zcs = self.Zcs
        omegas = self.omegas
        CASs = self.CASs
        return [Stockmayer(Tm=Tms[i], Tb=Tbs[i], Tc=Tcs[i], 
                           Zc=Zcs[i], omega=omegas[i], CASRN=CASs[i]) 
                for i in self.cmps]
    @CachedProperty
    def molecular_diameters(self):
        Tcs = self.Tcs
        Zcs = self.Zcs
        omegas = self.omegas
        CASs = self.CASs
        Vml_Tms = self.Vml_Tms
        Vml_Tb = self.Vml_Tb
        Pcs = self.Pcs
        Vcs = self.Vcs
        return [molecular_diameter(Tc=Tcs[i], Pc=Pcs[i], Vc=Vcs[i],
                                   Zc=Zcs[i], omega=omegas[i],
                                   Vm=Vml_Tms[i], Vb=Vml_Tb[i], 
                                   CASRN=CASs[i])
                for i in self.cmps]
    @CachedProperty
    def Psat_298s(self):
        return [i(298.15) for i in self.chemicals.VaporPressures]
    @CachedProperty
    def Hvap_298s(self):
        return [o.T_dependent_property(298.15) for o in self.chemicals.EnthalpyVaporizations]
    @CachedProperty
    def Hvap_298s_mass(self):
        return [Hvap*1000.0/MW if Hvap is not None else None for Hvap, MW in zip(self.Hvap_298s, self.MWs)]
    @CachedProperty
    def Vms_Tms(self):
        return [None if Tm is None else i.T_dependent_property(Tm) for i, Tm in zip(self.VolumeSolids, self.Tms)]
    @CachedProperty
    def rhos_Tms(self):
        return [1.0/V if V is not None else None for V in self.chemicals.Vms_Tms]
    @CachedProperty
    def rhos_Tms_mass(self):
        return [1e-3*MW/V if V is not None else None for V, MW in zip(self.Vms_Tms, self.MWs)]
    @CachedProperty
    def sigma_STPs(self):
        return [i.T_dependent_property(298.15) for i in self.SurfaceTensions]
    @CachedProperty
    def sigma_Tbs(self):
        return [None if Tb is None else i.T_dependent_property(Tb)
                for i, Tb in zip(self.SurfaceTensions, self.Tbs)]
    @CachedProperty
    def sigma_Tms(self):
        return [None if Tm is None else i.T_dependent_property(Tm)
               for i, Tm in zip(self.SurfaceTensions, self.Tms)]
    @CachedProperty
    def Van_der_Waals_volumes(self):
        N = self.N
        N_range = self.cmps
        UNIFAC_Rs = [None]*N
        UNIFAC_groups = self.UNIFAC_groups
        for i in N_range:
            groups = UNIFAC_groups[i]
            if groups is not None:
                UNIFAC_Rs[i] = UNIFAC_RQ(groups)[0]
        return [Van_der_Waals_volume(UNIFAC_Rs[i]) if UNIFAC_Rs[i] is not None else None for i in N_range]
    @CachedProperty
    def Van_der_Waals_areas(self):
        N = self.N
        N_range = self.cmps
        UNIFAC_Qs = [None]*N
        UNIFAC_groups = self.UNIFAC_groups
        for i in N_range:
            groups = UNIFAC_groups[i]
            if groups is not None:
                UNIFAC_Qs[i] = UNIFAC_RQ(groups)[1]
        return [Van_der_Waals_area(UNIFAC_Qs[i]) if UNIFAC_Qs[i] is not None else None for i in N_range]
    @CachedProperty
    def Parachors(self):
        N = self.N
        Parachors = [None]*N
        sigma_STPs = self.sigma_STPs
        MWs = self.MWs
        rhol_STPs_mass = self.rhol_STPs_mass
        rhog_STPs_mass = self.rhog_STPs_mass
        for i in self.cmps:
            try:
                Parachors[i] = Parachor(sigma=sigma_STPs[i], MW=MWs[i], rhol=rhol_STPs_mass[i], rhog=rhog_STPs_mass[i])
            except:
                pass
        return Parachors
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module extends the reaction module from the chemicals library:
# https://github.com/CalebBell/chemicals
# Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>
#
# This module is under a dual license:
# 1. The UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
# 
# 2. The MIT open-source license. See
# https://github.com/CalebBell/chemicals/blob/master/LICENSE.txt for details.

import os
from chemicals import reaction
import pandas as pd

reaction.__all__.extend([
    'Hf', 'S0', 'free_energy_at_phase'
])

folder = os.path.join(os.path.dirname(__file__), 'Reaction')
Hf_biochemicals = pd.read_csv(os.path.join(folder, 'Biochemicals Hf.tsv'), sep='\t', index_col=0)

def Hf(CASRN, phase=None, Hvap=None, Hfus=None):
    r'''
    Return a chemical's heat of formation at given phase. If no phase given,
    return the standard-state heat of formation. Return None if the data is 
    not available. 
    
    Parameters
    ----------
    CASRN : string
        CASRN [-].
    
    Returns
    -------
    Hf : float
        Standard-state heat of formation [J/mol].
    
    Other Parameters
    ----------------
    phase : 's', 'l', or 'g', optional
        Phase.
    Hvap=None : float, optional
        Heat of vaporization [J/mol].
    Hfus=None : float, optional
        Heat of fusion [J/mol].
    
    Notes
    -----
    This is a wrapper around :func:`~chemicals.reaction.Hfs`,
    :func:`~chemicals.reaction.Hfl`, and :func:`~chemicals.reaction.Hfg`.
    
    Examples
    --------
    >>> Hf(CASRN='7732-18-5')
    -241822.0
    
    '''
    if CASRN in Hf_biochemicals.index:
        Hf = Hf_biochemicals.at[CASRN, 'Hf']
        phase_found = Hf_biochemicals.at[CASRN, 'phase']
        if phase and phase_found != phase: 
            Hf = free_energy_at_phase(Hf, phase_found, phase, Hvap, Hfus)
        return Hf
    if not phase:
        for f in (reaction.Hfg, reaction.Hfl, reaction.Hfs):
            Hf = f(CASRN)
            if Hf: break 
    elif phase == 'g':
        Hf = reaction.Hfg(CASRN)
    elif phase == 'l':
        Hf = reaction.Hfl(CASRN)
    elif phase == 's':
        Hf = reaction.Hfs(CASRN)
    else:
        raise ValueError("phase must be either 's', 'l', or 'g'; not '%s'" %phase)
    if Hf is None:
        if phase != 'g':
            Hf = free_energy_at_phase(reaction.Hfg(CASRN), 'g', phase, Hvap, Hfus)
        elif phase != 'l':
            Hf = free_energy_at_phase(reaction.Hfl(CASRN), 'l', phase, Hvap, Hfus)
        elif phase != 's':
            Hf = free_energy_at_phase(reaction.Hfs(CASRN), 's', phase, Hvap, Hfus)
    return Hf
reaction.Hf = Hf

def S0(CASRN, phase=None, Svap=None, Sfus=None):
    r'''
    Return a chemical's absolute entropy of formation at given phase. If no phase given,
    return the standard-state entropy of formation. Return None if the data is 
    not available. 
    
    Parameters
    ----------
    CASRN : string
        CASRN [-].
    
    Returns
    -------
    S0 : float
        Absolute entropy of formation [J/mol].
    
    Other Parameters
    ----------------
    phase : 's', 'l', or 'g', optional
        Phase.
    Svap=None : float, optional
        Entropy of vaporization [J/mol].
    Sfus=None : float, optional
        Entropy of fusion [J/mol].
    
    Notes
    -----
    This is a wrapper around :func:`~chemicals.reaction.S0s`,
    :func:`~chemicals.reaction.S0l`, and :func:`~chemicals.reaction.S0g`.
    
    Examples
    --------
    >>> S0(CASRN='7732-18-5')
    188.8
    
    '''
    if not phase:
        for f in (reaction.S0g, reaction.S0l, reaction.S0s):
            S0 = f(CASRN)
            if S0: break 
    elif phase == 'g':
        S0 = reaction.S0g(CASRN)
    elif phase == 'l':
        S0 = reaction.S0l(CASRN)
    elif phase == 's':
        S0 = reaction.S0s(CASRN)
    else:
        raise ValueError("phase must be either 's', 'l', or 'g'; not '%s'" %phase)
    if S0 is None:
        if phase != 'g':
            S0 = free_energy_at_phase(reaction.S0g(CASRN), 'g', phase, Svap, Sfus)
        elif phase != 'l':
            S0 = free_energy_at_phase(reaction.S0l(CASRN), 'l', phase, Svap, Sfus)
        elif phase != 's':
            S0 = free_energy_at_phase(reaction.S0s(CASRN), 's', phase, Svap, Sfus)
    return S0
reaction.S0 = S0

def free_energy_at_phase(Ef, phase_ref, phase, Evap, Efus):
    r"""
    Return a chemical's free energy at given phase.
    Return None if value cannot be computed.
    
    Parameters
    ----------
    Ef : float
        Standard-state free energy of formation [J/mol].
    phase : 's', 'l', or 'g'
        Phase.
    Evap=None : float, optional
        Free energy of vaporization [J/mol].
    Efus=None : float, optional
        Free energy of fusion [J/mol].
    
    Returns
    -------
    Ef : float
        Free energy of formation at given phase [J/mol].
    
    """
    if Ef is None: return Ef
    elif phase_ref == phase:
        return Ef
    elif phase_ref == 'g' and phase == 'l':
        if Evap: return Ef - Evap
    elif phase_ref == 'g' and phase == 's':
        if Evap and Efus is not None: return Ef - Evap - Efus
    elif phase_ref == 'l' and phase == 'g':
        if Evap: return Ef + Evap
    elif phase_ref == 'l' and phase == 's':
        if Efus is not None: return Ef - Efus
    elif phase_ref == 's' and phase == 'l':
        if Efus is not None: return Ef + Efus
    elif phase_ref == 's' and phase == 'g':
        if Evap and Efus is not None: return Ef + Evap + Efus
    else: 
        raise ValueError("phase must be either 's', 'l', or 'g'; not '%s'" %phase)
reaction.free_energy_at_phase = free_energy_at_phase


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import os
import numba
from numba.experimental import jitclass
from numba.extending import as_numba_type

__all__ = ('jitdata',)

class MockJitclassAttribute:
    def __getattr__(self, name): pass
    
    def __setattr__(self, name, value): pass

mock_jitclass_attribute = MockJitclassAttribute()
mock_jitclass_attribute.instance_type = None

def jitdata(name_or_cls=None, /, py=False, **fields):
    """
    Create a jitclass with a lightweight __init__ method
    which saves input arguments as attributes.
        
    """
    if name_or_cls is None: return lambda name_or_cls: jitdata(name_or_cls, py=py, **fields)
    for i, j in fields.items():
        if isinstance(j, str): fields[i] = eval(j, numba.__dict__)
    if isinstance(name_or_cls, str):
        name = name_or_cls
        cls = None
    elif isinstance(name_or_cls, type):
        cls = name_or_cls
        name = cls.__name__
        annotations = cls.__annotations__
        for i, j in annotations.items():
            if isinstance(j, str): annotations[i] = eval(j, numba.__dict__)
        all_fields = {**fields, **cls.__annotations__}
    else:
        raise TypeError('first argument must be a class or name')
    if not all_fields: raise ValueError('at least one field must be given')
    if cls is not None and '__init__' in cls.__dict__:
        if py or os.environ.get("NUMBA_DISABLE_JIT"): 
            cls.class_type = mock_jitclass_attribute
            return cls
        return jitclass(cls, spec=[(i, as_numba_type(j)) for i, j in fields.items()])
    else:
        arguments = ', '.join(all_fields)
        init = [f"def __init__(self, {arguments}):"]
        for i in all_fields: init.append(f'self.{i} = {i}')
        init = '\n    '.join(init)
        dct = {}
        exec(init, dct)
        if cls is None: 
            cls = type(name, (), {'__init__': dct['__init__']})
        else:
            cls.__init__ = dct['__init__']
        # return cls
        if py or os.environ.get("NUMBA_DISABLE_JIT"): 
            cls.class_type = mock_jitclass_attribute
            return cls
        return jitclass(cls, spec=[(i, as_numba_type(j)) for i, j in fields.items()])

